-- @script Arizona Interface Extender
-- @author NyashMyash99 (nyashmyash99.ru)
-- @license https://files.foxinbox.su/license/LICENSE

--  
require 'moonloader'
local lib = require 'nmlibrary'
local keys = require 'vkeys'


-- 
script_name('Arizona Interface Extender')
script_description('      .')
script_version('1720615711')
script_author('NyashMyash99 (nyashmyash99.ru)')


-- 
local requiredLibraryVersions = { [1] = 1714121217, [2] = 1714121217 }
local scriptInformationUrl = 'https://arizona.nyashmyash99.ru/files/interfaceextender/update.json'

local configDirPath = getWorkingDirectory() .. '\\config\\ArizonaInterfaceExtender'
local configFilePath = configDirPath .. '\\config.ini'
local config = {}

local playerData = {
    level = 0,
    exp = 0,
    nextExp = 0,
    money = 0,
    bankMoney = 0,
    personalAccountMoney1 = 0,
    personalAccountMoney2 = 0,
    personalAccountMoney3 = 0,
    personalAccountMoney4 = 0,
    personalAccountMoney5 = 0,
    personalAccountMoney6 = 0,
    depositMoney = 0,
    azCoins = 0
}

local nextDataUpdate = -1

local isScriptLoaded = false
local isDebugMode = false


function main()
    sendDebug('[main] Waiting for game client to be ready.')
    while not isSampAvailable() do wait(500) end
    sendDebug('[main] Game client is ready, start initializing script.')

    sendMessage('')
    sendMessage(tostring(script.this.name) .. '  ' .. table.concat(script.this.authors, ', ') .. '  ' .. tostring(script.this.version) .. '.')
    sendMessage(script.this.description)
    sendMessage('')

    -- ,           :D
    if isDebugMode then
        sendMessage('         !', 'WARNING')
        sendMessage('')
        sendMessage('       ,    ...', 'WARNING')
        sendMessage('...  "local isDebugMode = true"  "local isDebugMode = false".', 'WARNING')
        sendMessage('')
        sendMessage('   : ' .. tostring(script.this.path) .. '.', 'WARNING')
    end

    --      -   .
    if not lib.connectLibrary(requiredLibraryVersions, isDebugMode) then
        return wait(-1)
    end

    --   .
    lib.checkUpdates(
        script.this.version,
        scriptInformationUrl,
        function (request)
            if not request.isSuccess then
                sendMessage(tostring(request.error), 'ERROR')
                sendMessage('      -    (nyashmyash99.ru).', 'ERROR')
                sendMessage('')
                sendMessage('*     -     ru ,   .', 'ERROR')
            end

            if request.isUpdateFound then
                sendMessage('    (' .. tostring(request.serverScriptVersion) .. '),     (BlastHack, Discord).', 'SUCCESS')
            end

            local isConfigurationReloaded = reloadConfiguration()
            --     -        .
            if not isConfigurationReloaded then return end

            registerCommands()
            registerListeners()
        end
    )

    while true do
        if not isScriptLoaded then goto continue end
        -- check hotkeys and execute them
        processHotkeys()

        -- update player data if it's time to update
        if nextDataUpdate ~= -1 and nextDataUpdate <= os.time() then
            -- temporarily disable update so it doesn't run bunch of times
            nextDataUpdate = -1
            updatePlayerData()
        end

        ::continue::
        wait(500)
    end
end


-- GUI
--- Inserts new elements into server interface.
function initializeInterface()
    sendDebug('[initializeInterface] Sending interface initialization request.')
    lib.executeJavaScript(string.format([[setTimeout(() => {
        // clean up old script styles
        document.querySelector('style#iestyles')?.remove();
        document.querySelector('link#ieicons')?.remove();

        // create style element
        var styles = document.createElement('style');
        document.body.append(styles);

        // add styles and icons
        styles.outerHTML = `
            <style id='iestyles'>
                .player-info {
                    width: 50vw;
                    align-items: flex-end;
                }

                .player-info__params {
                    position: relative;
                    transform: translateX(-30%%);
                }

                .player-info__player-money {
                    display: grid;
                    grid: repeat(2, 1fr) / repeat(3, 1fr);
                    grid-auto-flow: column;
                    column-gap: calc(.5vw + .5vh);

                    margin-top: calc(.5vw + .5vh);
                    margin-right: calc(1vw + 1vh);
                }

                .player-info__money-group {
                    width: 100%%;
                    display: flex;
                    align-items: center;
                }

                .player-info__fa-icon {
                    background: none;
                }

                .fa-solid, .fa-brands {
                    color: #68ff65;
                }

                .hide {
                    display: none;
                }
            </style>

            <link id='ieicons' rel='stylesheet' href='https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.4.2/css/all.min.css' integrity='sha512-z3gLpd7yknf1YoNbCzqRKc4qyor8gaKU1qmn+CShxbuBusANI9QpRohGBreCFkKxLhei6S9CQXFEbbKuqLg0DA==' crossorigin='anonymous' referrerpolicy='no-referrer' />
        `;


        // get server info bar
        var infoBarElement = document.querySelector('.player-info__server-info-bar');
        // clean up old leveling element
        infoBarElement.querySelector('.player-info__user-leveling')?.remove();

        // create leveling element
        var levelingElement = document.createElement('div');
        infoBarElement.prepend(levelingElement);

        levelingElement.outerHTML = `
            <div class='player-info__user-leveling player-info__user-stats'>
                <div class='player-info__user-level player-info__user-id'>
                    <span class='player-info__user-id-caption'>LVL:</span>
                    <span class='player-info__user-level-value player-info__user-id-value'>%s</span>
                </div>

                <div class='player-info__user-exp player-info__user-id'>
                    <span class='player-info__user-id-caption'>EXP:</span>
                    <span class='player-info__user-exp-value player-info__user-id-value'>%s</span>
                </div>
            </div>
        `;

        // get online section margin bottom
        var onlineMarginBottom = window.getComputedStyle(infoBarElement.querySelector('.player-info__users-online'))
            .getPropertyValue('margin-bottom');

        // modify leveling element
        infoBarElement.querySelector('.player-info__user-level').style.marginBottom = onlineMarginBottom;


        // modify money section
        var isBankMoneyVisible = %s;
        var isPersonalAccountMoneyVisible = %s;
        var isDepositMoneyVisible = %s;
        var isSummaryMoneyVisible = %s;
        var isAZCoinsVisible = %s;

        document.querySelector('.player-info__player-money').outerHTML = `
            <div class='player-info__player-money'>
                <span class='player-info__cash-money-group player-info__money-group'>
                    <span class='player-info__dollar-icon'>
                        <span class='player-info__symbol'>&nbsp;$&nbsp;</span>
                    </span>

                    <span class='player-info__total'>%s</span>
                </span>

                <span class='player-info__bank-money-group player-info__money-group ${isBankMoneyVisible ? '' : 'hide'}'>
                    <span class='player-info__fa-icon player-info__dollar-icon'>
                        <i class='player-info__symbol fa-solid fa-building-columns'></i>
                    </span>

                    <span class='player-info__bank player-info__total'>%s</span>
                </span>

                <span class='player-info__personal-account-money-group player-info__money-group ${isPersonalAccountMoneyVisible ? '' : 'hide'}'>
                    <span class='player-info__fa-icon player-info__dollar-icon'>
                        <i class='player-info__symbol fa-solid fa-wallet'></i>
                    </span>

                    <span class='player-info__personal-account player-info__total'>%s</span>
                </span>

                <span class='player-info__deposit-money-group player-info__money-group ${isDepositMoneyVisible ? '' : 'hide'}'>
                    <span class='player-info__fa-icon player-info__dollar-icon'>
                        <i class='player-info__symbol fa-solid fa-piggy-bank'></i>
                    </span>

                    <span class='player-info__deposit player-info__total'>%s</span>
                </span>

                <span class='player-info__summary-money-group player-info__money-group ${isSummaryMoneyVisible ? '' : 'hide'}'}>
                    <span class='player-info__fa-icon player-info__dollar-icon'>
                        <i class='player-info__symbol fa-solid fa-suitcase'></i>
                    </span>

                    <span class='player-info__summary player-info__total'>%s</span>
                </span>

                <span class='player-info__az-coins-group player-info__money-group ${isAZCoinsVisible ? '' : 'hide'}'}>
                    <span class='player-info__fa-icon player-info__dollar-icon'>
                        <i class='player-info__symbol fa-brands fa-adn'></i>
                    </span>

                    <span class='player-info__az-coins player-info__total'>%s</span>
                </span>
            </div>
        `;

        // get money icon font size
        var iconFontSize = window.getComputedStyle(document.querySelector('.player-info__symbol'))
            .getPropertyValue('font-size');

        // modify custom icons font size
        [...document.querySelectorAll('.fa-solid'), ...document.querySelectorAll('.fa-brands')]
            .forEach(icon => icon.style.fontSize = `calc(${iconFontSize} * 1.3)`);
    });]],
    tostring(playerData.level),
    tostring(playerData.exp) .. '/' .. tostring(playerData.nextExp),
    config.visibility.bankMoney,
    config.visibility.personalAccountMoney,
    config.visibility.depositMoney,
    config.visibility.summaryMoney,
    config.visibility.azCoins,
    lib.toMoneyFormat(playerData.money),
    lib.toMoneyFormat(playerData.bankMoney),
    lib.toMoneyFormat(playerData.personalAccountMoney1 + playerData.personalAccountMoney2 + playerData.personalAccountMoney3),
    lib.toMoneyFormat(playerData.depositMoney),
    lib.toMoneyFormat(calculateSummaryMoney()),
    lib.toMoneyFormat(playerData.azCoins)
    ))
    sendDebug('[initializeInterface] Interface initialization request sent.')
end

--- Render interface with player data.
function renderInterface()
    sendDebug('[renderInterface] Sending interface changing request: ' .. tostring(encodeJson(playerData)) .. '.')
    lib.executeJavaScript(string.format([[
        document.querySelector('.player-info__user-level-value').innerHTML = '%s';
        document.querySelector('.player-info__user-exp-value').innerHTML = '%s';
        document.querySelector('.player-info__total').innerHTML = '%s';
        document.querySelector('.player-info__bank').innerHTML = '%s';
        document.querySelector('.player-info__personal-account').innerHTML = '%s';
        document.querySelector('.player-info__deposit').innerHTML = '%s';
        document.querySelector('.player-info__summary').innerHTML = '%s';
        document.querySelector('.player-info__az-coins').innerHTML = '%s';

        // change visibility of summary money element
        var isBankMoneyVisible = %s;
        document.querySelector('.player-info__bank-money-group')
            .classList
            .toggle('hide', !isBankMoneyVisible);

        // change visibility of summary money element
        var isPersonalAccountMoneyVisible = %s;
        document.querySelector('.player-info__personal-account-money-group')
            .classList
            .toggle('hide', !isPersonalAccountMoneyVisible);

        // change visibility of summary money element
        var isDepositMoneyVisible = %s;
        document.querySelector('.player-info__deposit-money-group')
            .classList
            .toggle('hide', !isDepositMoneyVisible);

        // change visibility of summary money element
        var isSummaryMoneyVisible = %s;
        document.querySelector('.player-info__summary-money-group')
            .classList
            .toggle('hide', !isSummaryMoneyVisible);

        // change visibility of summary money element
        var isAZCoinsVisible = %s;
        document.querySelector('.player-info__az-coins-group')
            .classList
            .toggle('hide', !isAZCoinsVisible);
    ]],
    tostring(playerData.level),
    tostring(playerData.exp) .. '/' .. tostring(playerData.nextExp),
    lib.toMoneyFormat(playerData.money),
    lib.toMoneyFormat(playerData.bankMoney),
    lib.toMoneyFormat(playerData.personalAccountMoney1 + playerData.personalAccountMoney2 + playerData.personalAccountMoney3 + playerData.personalAccountMoney4 + playerData.personalAccountMoney5 + playerData.personalAccountMoney6),
    lib.toMoneyFormat(playerData.depositMoney),
    lib.toMoneyFormat(calculateSummaryMoney()),
    lib.toMoneyFormat(playerData.azCoins),
    config.visibility.bankMoney,
    config.visibility.personalAccountMoney,
    config.visibility.depositMoney,
    config.visibility.summaryMoney,
    config.visibility.azCoins
    ))
    sendDebug('[renderInterface] Interface changing request sent: ' .. tostring(encodeJson(playerData)) .. '.')
end

--- Calculates total balance depending on displayed items.
function calculateSummaryMoney()
    local summaryMoney = playerData.money

    if config.visibility.bankMoney then
        summaryMoney = summaryMoney + playerData.bankMoney
    end

    if config.visibility.personalAccountMoney then
        summaryMoney = summaryMoney + playerData.personalAccountMoney1 + playerData.personalAccountMoney2 + playerData.personalAccountMoney3 + playerData.personalAccountMoney4 + playerData.personalAccountMoney5 + playerData.personalAccountMoney6
    end

    if config.visibility.depositMoney then
        summaryMoney = summaryMoney + playerData.depositMoney
    end

    return summaryMoney
end


-- Functions
function registerListeners()
    lib.subscribeToDialog(
        lib.Dialogues.MAIN_STATISTIC,
        function(content, _, id)
            local foundedItems = {}

            -- run through all lines of dialog
            for _, value in ipairs(content) do
                sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Processing value: ' .. tostring(value) .. '.')

                if not foundedItems.azCoins then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching az coins.')
                    local azCoinsMatch = value:match('(%d+) AZ%-Coins')

                    if azCoinsMatch then
                        local azCoins = tonumber(azCoinsMatch)
                        playerData.azCoins = azCoins
                        foundedItems.azCoins = true

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found az coins: ' .. tostring(azCoins) .. '.')
                        goto nextLine
                    end
                end

                if not foundedItems.level and foundedItems.azCoins then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching level.')
                    local levelMatch = value:match(': %[(%d+)]')

                    if levelMatch then
                        local level = tonumber(levelMatch)
                        playerData.level = level
                        foundedItems.level = true

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found level: ' .. tostring(level) .. '.')
                        goto nextLine
                    end
                end

                if not foundedItems.exp and foundedItems.level then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching respect.')
                    local expMatch, nextExpMatch = value:match(': %[(%d+)/(%d+)]')

                    if expMatch and nextExpMatch then
                        local exp = tonumber(expMatch)
                        local nextExp = tonumber(nextExpMatch)

                        playerData.exp = exp
                        playerData.nextExp = nextExp
                        foundedItems.exp = true

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found respect: ' .. tostring(exp) .. '/' .. tostring(nextExp) .. '.')
                        goto nextLine
                    end
                end

                if not foundedItems.money and foundedItems.exp then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching money.')
                    local moneyMatch = value:match('  %(SA%$%): %[%$([%d,]+)]')

                    if moneyMatch then
                        moneyMatch = moneyMatch:gsub(',', '')

                        local money = tonumber(moneyMatch)
                        playerData.money = money
                        foundedItems.money = true

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found money: ' .. tostring(money) .. '.')
                        goto nextLine
                    end
                end

                if not foundedItems.personalAccountMoney then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching personal account money.')
                    local personalAccountNumberMatch, personalAccountMoneyMatch = value:match('   (%d): %[%$([%d,]+)]')

                    if personalAccountNumberMatch and personalAccountMoneyMatch then
                        personalAccountMoneyMatch = personalAccountMoneyMatch:gsub(',', '')

                        local personalAccountNumber = tonumber(personalAccountNumberMatch)
                        local money = tonumber(personalAccountMoneyMatch)

                        playerData['personalAccountMoney' .. tostring(personalAccountNumber)] = money
                        if personalAccountNumber == 6 then
                            foundedItems.personalAccountMoney = true
                        end

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found personal account money (' .. tostring(personalAccountNumber) .. '): ' .. tostring(money) .. '.')
                        goto nextLine
                    end
                end

                if not foundedItems.bankMoney and foundedItems.money then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching bank money.')
                    local bankMoneyMatch = value:match('  : %[%$([%d,]+)]')

                    if bankMoneyMatch then
                        bankMoneyMatch = bankMoneyMatch:gsub(',', '')

                        local money = tonumber(bankMoneyMatch)
                        playerData.bankMoney = money
                        foundedItems.bankMoney = true

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found bank money: ' .. tostring(money) .. '.')
                        goto nextLine
                    end
                end

                if not foundedItems.depositMoney and foundedItems.bankMoney then
                    sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Matching deposit money.')
                    local depositMoneyMatch = value:match('  : %[%$([%d,]+)]')

                    if depositMoneyMatch then
                        depositMoneyMatch = depositMoneyMatch:gsub(',', '')

                        local money = tonumber(depositMoneyMatch)
                        playerData.depositMoney = money
                        foundedItems.depositMoney = true

                        sendDebug('[MAIN_STATISTIC] (' .. tostring(id) .. ') Found deposit money: ' .. tostring(money) .. '.')
                    end
                end

                ::nextLine::
            end

            renderInterface()
        end
    )

    --- Registers listener that checks if interface needs to be initialized.
    lib.subscribeToIncomingPacket(
        lib.Packets.CLIENT_EXECUTE_JAVASCRIPT,
        function (event, view)
            if event ~= 'event.setActiveView' then return end

            sendDebug('[ServerInterfaceListener] Active view: ' .. tostring(view) .. '.')
            if view ~= ' null ' then return end

            sendDebug('[ServerInterfaceListener] View changed to Arizona HUD, reinitializing interface.')
            initializeInterface()

            if not isScriptLoaded then
                -- mark script as loaded to start processing hotkeys
                isScriptLoaded = true
                nextDataUpdate = os.time() + 1

                sendDebug('[ServerInterfaceListener] Script is fully initialized.')
            end
        end
    )
end

function updatePlayerData()
    sendDebug('[updatePlayerData] Starting collecting statistics.')
    if not openStatisticDialog() then return end

    nextDataUpdate = os.time() + tonumber(config.autoUpdate.period)
    sendDebug('[updatePlayerData] Waiting for next update task: ' .. tostring(nextDataUpdate) .. '.')
end

function openStatisticDialog()
    sendDebug('[openStatisticDialog] Discovery request received.')

    if sampIsDialogActive() then
        sendDebug('[openStatisticDialog] Another dialog is active, wait 1 second.')
        nextDataUpdate = os.time() + 1
        return false
    end

    -- mark dialog 'Main Statistic' as expected and enter command
    lib.expectDialog(lib.Dialogues.MAIN_STATISTIC)

    sendDebug('[openStatisticDialog] Dispatching command: /stats.')
    sampSendChat('/stats')
    sendDebug('[openStatisticDialog] Command dispatched: /stats.')

    return true
end


-- Configuration
function reloadConfiguration()
    local reloadingResult = lib.reloadConfiguration(
        configDirPath,
        configFilePath,
        {
            documentation = {
                link = 'https://arizona.nyashmyash99.ru/files/interfaceextender/documentation.md'
            },

            hotkeys = {
                forceUpdate = 'VK_NUMPAD0'
            },

            autoUpdate = {
                period = 10,
            },

            visibility = {
                bankMoney = true,
                personalAccountMoney = true,
                depositMoney = true,
                summaryMoney = true,
                azCoins = true
            }
        }
    )

    if not reloadingResult.isSuccess then
        sendMessage('{FF0000} ' .. tostring(reloadingResult.error))
        sendMessage('{FF0000} Restart script and contact developer (nyashmyash99.ru) if this error happens again.')
        return false
    end

    config = reloadingResult.config
    return true
end

function saveConfiguration()
    lib.saveConfiguration(
        configDirPath,
        configFilePath,
        config
    )
end


-- Commands
function registerCommands()
    sendDebug('[registerCommands] Registering "ie" command.')
    sampRegisterChatCommand('ie', interfaceExtenderCommand)
    sendDebug('[registerCommands] Command "ie" registered.')
end

--- @param text string
function interfaceExtenderCommand(text)
    sendDebug('[ie] Command dispatched: ' .. tostring(text) .. '.')
    local args = lib.split(text, ' ')
    local subCommand = args[1]
    sendDebug('[ie] Subcommand: ' .. tostring(subCommand) .. '.')

    if subCommand == 'visibility' or subCommand == 'v' then
        return visibilitySubCommand(args[2])
    end

    if subCommand == 'reload' or subCommand == 'r' then
        return reloadSubCommand()
    end

    sendDebug('[ie] No such subcommand found: ' .. tostring(subCommand) .. '.')
    sendMessage('Interface extender commands:')
    sendMessage('visibility [part] - enable/disable display of certain part of interface')
    sendMessage('reload - full script reloading without restarting game')
end

--- Enables/disables display of certain part of interface.
--- @param part string
function visibilitySubCommand(part)
    sendDebug('[ie visibility] Command dispatched: ' .. tostring(part) .. '.')
    local parts, part2Bool = lib.tableKeys(config.visibility)
    sendDebug('[ie visibility] Available parts: ' .. table.concat(parts, ', ') .. '.')

    if not part or not part2Bool[part] then
        sendDebug('[ie visibility] No such part found: ' .. tostring(part) .. '.')
        sendMessage('Components with ability to change visibility:')
        sendMessage(table.concat(parts, ', ') .. '.')
        return
    end

    local enabled = config.visibility[part]

    sendDebug('[ie visibility] Past "visibility.' .. tostring(part) .. '" value: ' .. tostring(enabled) .. '.')
    config.visibility[part] = not enabled
    sendDebug('[ie visibility] New "visibility.' .. tostring(part) .. '" value: ' .. tostring(config.visibility[part]) .. '.')

    saveConfiguration()
    renderInterface()
end

--- Reloads script.
function reloadSubCommand()
    sendDebug('[ie reload] Command dispatched.')
    script.this:reload()
end


-- Hotkey handlers
function processHotkeys()
    if sampIsCursorActive() then return end

    onForceUpdate()
end

function onForceUpdate()
    local key = config.hotkeys.forceUpdate

    if isKeyJustPressed(keys[key]) then
        sendDebug('[onForceUpdate] Key pressed: ' .. tostring(key) .. '.')

        sendDebug('[onForceUpdate] Force updating player data.')
        updatePlayerData()
    end
end


--   
---        .
--- @param message string
--- @param type string INFO/SUCCESS/WARNING/ERROR (  INFO)
function sendMessage(message, type)
    local color
    if type == 'SUCCESS' or type == 'WARNING' or type == 'ERROR' then
        color = lib.Colors[type]
    else
        color = '{FFA500}'
    end

    sampAddChatMessage(' > ' .. tostring(color) .. tostring(message), 0xFFA500)
end

---     ,    .
--- @param message string
function sendDebug(message)
    if not isDebugMode then return end
    print(tostring(message))
end