---       
--- @module NMLibrary
--- @author NyashMyash99 (nyashmyash99.ru)
--- @license https://files.foxinbox.su/license/LICENSE

-- ==   ==
local sampev = require 'lib.samp.events'
local inicfg = require 'inicfg'
local effil = require 'effil'

local encoding = require 'encoding'
encoding.default = 'CP1251'


-- ==  ==
-- ===  ===
--- ID  .
local Packets = {
    ---  JS    .
    CLIENT_EXECUTE_JAVASCRIPT = 17,
    ---       .
    SERVER_EXECUTE_EVENT = 18
}

local incomingPacketListeners = {}
local outgoingPacketListeners = {}

-- ===  ===
--- ID  .
local Dialogues = {
    --- /mn
    MENU = 722,
    --- /stats
    MAIN_STATISTIC = 235,
    --- /donate
    DONATE = 222,
    --- /quest
    QUESTS = 25854,
    ---  / 
    QUEST = 7653,
    --- /lmenu
    LEADER_MENU = 1214,
    ---   /  
    ORGANIZATION_MEMBERS_TYPE = 1009,
    --- /members
    ORGANIZATION_ONLINE_MEMBERS = 2015,
    ORGANIZATION_OFFLINE_MEMBERS = 567,
    --- /invite
    INVITE_RANK = 25636,
    --- /fmembers |     /   
    FAMILY_ONLINE_MEMBERS_TYPE = 25526,
    FAMILY_ONLINE_MEMBERS = 1488,
    FAMILY_OFFLINE_MEMBERS = 2931,
    --- /mn -> [14]  
    CORRECTIONAL_WORK = 0
}

local dialogListeners = {}
--- ,  .
local expectedDialogues = {}

-- ===  ===
---  OOC ,    .
local ChatTypes = {
    LOCAL = 1,
    FAMILY = 2,
    LEGAL_ORGANIZATION = 3,
    ILLEGAL_ORGANIZATION = 4
}

local commands = {}

-- ===  ===
local Colors = {
    ERROR = '{FF0000}',
    WARNING = '{FF4500}',
    SUCCESS = '{00FF00}'
}


-- ==  ==
local NMLibrary = {
    _VERSION = '1714121217',
    debug = {
        ---    .
        global = false,
        PACKET = {
            ---       .
            BYTES = false,
            ---     .
            INCOMING = false,
            ---     .
            OUTGOING = false
        }
    },
    Packets = Packets,
    Dialogues = Dialogues,
    Colors = Colors,
    ChatTypes = ChatTypes
}


-- ==    ==
---     ,    .
--- @param message string
--- @param debugMode boolean  ,         (    )
local function sendDebug(message, debugMode)
    if debugMode == nil then debugMode = NMLibrary.debug.global end
    if not debugMode then return end
    print(tostring(message))
end


-- ==  ==
--- @param requiredLibraryVersions table  { [1] =   , [2] =    }
--- @param debugMode boolean  ,         (    )
--- @return boolean     
function NMLibrary.connectLibrary(requiredLibraryVersions, debugMode)
    if debugMode then
        NMLibrary.debug.global = debugMode
        sendDebug('[connectLibrary]   .')
    end

    sendDebug('[connectLibrary]    : ' .. tostring(encodeJson(requiredLibraryVersions)) .. '.')
    local minRequiredVersion = requiredLibraryVersions[1]
    local maxRequiredVersion = requiredLibraryVersions[2]

    if tonumber(minRequiredVersion) > tonumber(NMLibrary._VERSION) then
        sendDebug('[connectLibrary])    : ' .. tostring(NMLibrary._VERSION) .. ' (), ' .. tostring(minRequiredVersion) .. ' (. ).')
        sendMessage(Colors.ERROR .. '     NMLibrary,     (BlastHack, Discord).')
        return false
    end

    if tonumber(maxRequiredVersion) < tonumber(NMLibrary._VERSION) then
        sendDebug('[connectLibrary]    : ' .. tostring(NMLibrary._VERSION) .. ' (), ' .. tostring(maxRequiredVersion) .. ' (. ).')
        sendMessage(Colors.ERROR .. '     NMLibrary,      ,    .')
        return false
    end

    return true
end


-- ==  ==
---  ,    - .
function sampev.onShowDialog(id, _, title, _, _, text)
    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')

    if not expectedDialogues[id] then
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')
        return true
    end

    NMLibrary.releaseDialog(id)


    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')
    local listeners = dialogListeners[id]

    if not listeners then
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')
        return true
    end
    sendDebug('[onShowDialog] (' .. tostring(id) .. ')  ' .. tostring(#listeners) .. ' .')


    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')
    title = NMLibrary.trim(NMLibrary.stripColors(title))
    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   : ' .. tostring(title) .. '.')

    sendDebug('[onShowDialog] (' .. tostring(id) .. ')     .')
    text = NMLibrary.split(text, '\n')
    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   : ' .. tostring(encodeJson(text)) .. '.')

    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')
    for i, line in ipairs(text) do
        text[i] = NMLibrary.trim(NMLibrary.stripColors(line))
    end
    sendDebug('[onShowDialog] (' .. tostring(id) .. ')   : ' .. tostring(encodeJson(text)) .. '.')


    local isNeedToBeClosed = true
    local isNeedToBeDisplayed = false
    for i, listener in ipairs(listeners) do
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')  ' .. tostring(i) .. ' .')
        local listenerClosingResponse, listenerDisplayResponse = listener(text, title, id)
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')  #' .. tostring(i) .. ' : ' .. tostring(listenerClosingResponse) .. ', ' .. tostring(listenerDisplayResponse) .. '.')

        --  ,      .
        if listenerClosingResponse == false then
            isNeedToBeClosed = false
        end

        --  ,     .
        if listenerDisplayResponse == true then
            isNeedToBeDisplayed = true
        end
    end

    --  ,   .
    if isNeedToBeClosed then
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')      .')
        sampSendDialogResponse(id, 0, 0, nil)
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')      .')
    else
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')     .')
    end

    if isNeedToBeDisplayed then
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')     .')
    else
        sendDebug('[onShowDialog] (' .. tostring(id) .. ')   .')
    end

    return isNeedToBeDisplayed
end

--- @param id number
--- @param listener function    (string[])    (string),              (boolean, boolean) (  true, false)
function NMLibrary.subscribeToDialog(id, listener)
    if not dialogListeners[id] then
        sendDebug('[subscribeToDialog] (' .. tostring(id) .. ')    .')
        dialogListeners[id] = {}
        sendDebug('[subscribeToDialog] (' .. tostring(id) .. ')    .')
    end

    sendDebug('[subscribeToDialog] (' .. tostring(id) .. ')   .')
    table.insert(dialogListeners[id], listener)
    sendDebug('[subscribeToDialog] (' .. tostring(id) .. ')   .')
end

---   ,           ( ).
--- @param id number
function NMLibrary.expectDialog(id)
    sendDebug('[expectDialog]   "' .. tostring(id) .. '"  .')
    expectedDialogues[id] = true
    sendDebug('[expectDialog]   "' .. tostring(id) .. '": ' .. tostring(NMLibrary.tableSize(expectedDialogues)) .. '  .')
end

---   .
--- @param id number
function NMLibrary.releaseDialog(id)
    sendDebug('[releaseDialog]    "' .. tostring(id) .. '".')
    expectedDialogues[id] = nil
    sendDebug('[releaseDialog]   "' .. tostring(id) .. '" : ' .. tostring(NMLibrary.tableSize(expectedDialogues)) .. '  .')
end


-- ==  ==
---  -.
--- @param syntax string regex   ,        
--- @param chatType number  ,     (. ChatTypes)
--- @param handler function    commandRegex (...string),      (boolean) (  true)
--- @param cooldown number  ,      (  -1)
--- @param notifyErrors boolean  ,       ( - ) (  false)
function NMLibrary.registerCommand(
    syntax,
    chatType,
    handler,
    cooldown,
    notifyErrors
)
    local command = {
        syntax = syntax,
        chatType = chatType,
        cooldown = cooldown or -1,
        notifyErrors = notifyErrors or false
    }

    sendDebug('[registerCommand]   : ' .. encodeJson(command) .. '.')

    command.handler = handler
    table.insert(commands, command)

    sendDebug('[registerCommand]   :  ' .. tostring(#commands) .. ' .')
end

function NMLibrary.processCommand(text)
    for _, command in ipairs(commands) do
        local sender, senderId, senderRank, message, answerCommand

        if command.chatType == ChatTypes.LOCAL then
            local player, playerId, localMessage = text:match('%(%( (.-)%[(%d+)]: (.+) %)%)')
            if not player or not playerId or not localMessage then
                goto nextCommand
            end

            sender = player
            senderId = playerId
            message = localMessage
            answerCommand = '/b'
        end

        if command.chatType == ChatTypes.LEGAL_ORGANIZATION then
            local playerRank, player, playerId, organizationMessage = text:match('%[R] (.+) (.-)%[(%d+)]: %(%( (.+) %)%)')
            if not player or not playerRank or not playerId or not organizationMessage then
                goto nextCommand
            end

            sender = player
            senderId = playerId
            senderRank = playerRank
            message = organizationMessage
            answerCommand = '/rb'
        end

        if command.chatType == ChatTypes.ILLEGAL_ORGANIZATION then
            local playerRank, player, playerId, organizationMessage = text:match('%[F] (.+) (.-)%[(%d+)]: %(%( (.+) %)%)')
            if not player or not playerRank or not playerId or not organizationMessage then
                goto nextCommand
            end

            sender = player
            senderId = playerId
            senderRank = playerRank
            message = organizationMessage
            answerCommand = '/fb'
        end

        --      .
        local arguments = table.pack(message:match(command.syntax))
        if #arguments == 0 then
            goto nextCommand
        end

        if command.cooldownUntil ~= nil and command.cooldownUntil > os.time() then
            if not command.notifyErrors then
                goto nextCommand
            end

            sampSendChat(tostring(answerCommand) .. '        ' .. tostring(math.ceil((command.cooldownUntil - os.time()) / 60)) .. ' .')
            goto nextCommand
        end

        local isSuccess = command.handler(sender, senderId, senderRank, table.unpack(arguments))
        --     ,  false -   ,   ,   .
        if isSuccess ~= false and command.cooldown ~= -1 then
            command.cooldownUntil = os.time() + command.cooldown
        end

        ::nextCommand::
    end
end

---   .
function sampev.onServerMessage(_, text)
    text = NMLibrary.stripColors(text)
    NMLibrary.processCommand(text)
    return true
end


-- == CEF ==
---      JavaScript    .
--- @param code string
function NMLibrary.executeJavaScript(code)
    sendDebug('[executeJavaScript]  JavaScript : ' .. tostring(code) .. '.')

    local stream = raknetNewBitStream()

    --  ID .
    raknetBitStreamWriteInt8(stream, Packets.CLIENT_EXECUTE_JAVASCRIPT)
    --  -  .
    raknetBitStreamWriteInt32(stream, 0)
    --    .
    raknetBitStreamWriteInt32(stream, #code)
    --  .
    raknetBitStreamWriteString(stream, code)

    --   CEF .
    raknetEmulPacketReceiveBitStream(220, stream)

    --  .
    raknetDeleteBitStream(stream)

    sendDebug('[executeJavaScript] JavaScript  .')
end

---         .
--- @param event string
--- @param arguments string
function NMLibrary.executeClientEvent(event, arguments)
    NMLibrary.executeJavaScript(('window.executeEvent(\'%s\', \'[%s]\');'):format(event, arguments))
end

---        .
--- @param event string
--- @param arguments string
function NMLibrary.executeServerEvent(event, arguments)
    local action = tostring(event) .. '|' .. tostring(arguments)
    sendDebug('[executeServerEvent]  : ' .. tostring(action) .. '.')

    local stream = raknetNewBitStream()

    --   ID .
    raknetBitStreamWriteInt8(stream, 220)
    --  ID .
    raknetBitStreamWriteInt8(stream, 18)
    --    .
    raknetBitStreamWriteInt32(stream, #action)
    --  .
    raknetBitStreamWriteString(stream, action)

    --    .
    raknetSendBitStream(stream);

    --  .
    raknetDeleteBitStream(stream);

    sendDebug('[executeServerEvent]  : ' .. tostring(action) .. '.')
end

---    CEF .
--- @param packetID number
--- @param listener function
function NMLibrary.subscribeToIncomingPacket(packetID, listener)
    if not incomingPacketListeners[packetID] then
        sendDebug('[subscribeToIncomingPacket] (' .. tostring(packetID) .. ')    .')
        incomingPacketListeners[packetID] = {}
        sendDebug('[subscribeToIncomingPacket] (' .. tostring(packetID) .. ')    .')
    end

    sendDebug('[subscribeToIncomingPacket] (' .. tostring(packetID) .. ')   .')
    table.insert(incomingPacketListeners[packetID], listener)
    sendDebug('[subscribeToIncomingPacket] (' .. tostring(packetID) .. ')   .')
end

---    CEF .
--- @param packetID number
--- @param listener function
function NMLibrary.subscribeToOutgoingPacket(packetID, listener)
    if not outgoingPacketListeners[packetID] then
        sendDebug('[subscribeToOutgoingPacket] (' .. tostring(packetID) .. ')    .')
        outgoingPacketListeners[packetID] = {}
        sendDebug('[subscribeToOutgoingPacket] (' .. tostring(packetID) .. ')    .')
    end

    sendDebug('[subscribeToOutgoingPacket] (' .. tostring(packetID) .. ')   .')
    table.insert(outgoingPacketListeners[packetID], listener)
    sendDebug('[subscribeToOutgoingPacket] (' .. tostring(packetID) .. ')   .')
end

---  CEF      .
--- @param packetID number
--- @param stream table
--- @param type string incoming/outgoing
function NMLibrary.processPacket(packetID, stream, type)
    if packetID == Packets.CLIENT_EXECUTE_JAVASCRIPT then
        --   4  .
        raknetBitStreamIgnoreBits(stream, 32)

        --    .
        local codeLength = raknetBitStreamReadInt32(stream)
        if codeLength == 0 then return nil end

        --  .
        local code = raknetBitStreamReadString(stream, codeLength)

        --     .
        local event, arguments = code:match("window%.executeEvent%('(.+)', '%[(.+)]'%)")

        sendDebug('[processPacket]   ' .. tostring(NMLibrary.findKey(Packets, packetID)) .. ' (' .. tostring(event) .. '): ' .. tostring(arguments) .. '.', NMLibrary.debug.PACKET[type])
        return event, arguments
    end

    if packetID == Packets.SERVER_EXECUTE_EVENT then
        --    .
        local actionLength = raknetBitStreamReadInt32(stream)
        if actionLength == 0 then return nil end

        --  .
        local action = raknetBitStreamReadString(stream, actionLength)

        --     .
        local event, arguments = action:match("(.-)|(.*)")

        sendDebug('[processPacket]   ' .. tostring(NMLibrary.findKey(Packets, packetID)) .. ' (' .. tostring(event) .. '): ' .. tostring(arguments) .. '.', NMLibrary.debug.PACKET[type])
        return event, arguments
    end

    return nil
end

---  CEF .
addEventHandler('onReceivePacket', function(id, stream)
    if id ~= 220 then return end

    local bytes = {}
    --      .
    for _ = 1, raknetBitStreamGetNumberOfBytesUsed(stream) do
        table.insert(bytes, raknetBitStreamReadInt8(stream))
    end
    sendDebug('[onReceivePacket]  : ' .. tostring(table.concat(bytes, ', ')) .. '.', NMLibrary.debug.PACKET.INCOMING and NMLibrary.debug.PACKET.BYTES)

    --   .
    raknetBitStreamResetReadPointer(stream)
    --   ID .
    raknetBitStreamIgnoreBits(stream, 8)

    local packetID = raknetBitStreamReadInt8(stream)
    local listeners = incomingPacketListeners[packetID]
    if not listeners then return end

    --       .
    local packetData = table.pack(NMLibrary.processPacket(packetID, stream, 'incoming'))

    for _, listener in ipairs(listeners) do
        listener(table.unpack(packetData))
    end
end)

addEventHandler('onSendPacket', function(id, stream)
    if id ~= 220 then return end

    local bytes = {}
    --      .
    for _ = 1, raknetBitStreamGetNumberOfBytesUsed(stream) do
        table.insert(bytes, raknetBitStreamReadInt8(stream))
    end
    sendDebug('[onSendPacket]  : ' .. tostring(table.concat(bytes, ', ')) .. '.', NMLibrary.debug.PACKET.OUTGOING and NMLibrary.debug.PACKET.BYTES)

    --   .
    raknetBitStreamResetReadPointer(stream)
    --   ID .
    raknetBitStreamIgnoreBits(stream, 8)

    local packetID = raknetBitStreamReadInt8(stream)
    local listeners = outgoingPacketListeners[packetID]
    if not listeners then return end

    --       .
    local packetData = table.pack(NMLibrary.processPacket(packetID, stream, 'outgoing'))

    for _, listener in ipairs(listeners) do
        listener(table.unpack(packetData))
    end
end)


-- ==  ==
--- @param scriptVersion number
--- @param scriptInformationUrl string
--- @param callback function   { isSuccess: boolean, isUpdateFound?: boolean, serverScriptVersion?: number, error?: string }
function NMLibrary.checkUpdates(scriptVersion, scriptInformationUrl, callback)
    sendDebug('[checkUpdates]       : ' .. tostring(scriptInformationUrl) .. '.')
    NMLibrary.getRequest(scriptInformationUrl, function (statusCode, data, error)
        sendDebug('[checkUpdates]    : ' .. tostring(scriptInformationUrl) .. '.')

        if error then
            sendDebug('[checkUpdates] -        : ' .. tostring(error) .. ' (' .. tostring(scriptInformationUrl) .. ').')
            return callback({
                isSuccess = false,
                error = '-        .'
            })
        end

        if statusCode ~= 200 then
            sendDebug('[checkUpdates] -        : ' .. tostring(statusCode) .. '   (' .. tostring(scriptInformationUrl) .. ').')
            return callback({
                isSuccess = false,
                error = '-        .'
            })
        end

        sendDebug('[checkUpdates]  : ' .. tostring(encodeJson(data)) .. ' (' .. tostring(scriptInformationUrl) .. ').')
        local serverScriptVersion = data['version']

        --     -   .
        if tonumber(serverScriptVersion) <= tonumber(scriptVersion) then
            sendDebug('[checkUpdates]   : ' .. tostring(serverScriptVersion) .. ' ( ), ' ..  tostring(scriptVersion) .. ' () (' .. tostring(scriptInformationUrl) .. ').')
            return callback({
                isSuccess = true,
                isUpdateFound = false
            })
        end

        sendDebug('[checkUpdates]    : ' .. tostring(serverScriptVersion) .. ' (' .. tostring(scriptInformationUrl) .. ').')
        callback({
            isSuccess = true,
            isUpdateFound = false,
            serverScriptVersion = tonumber(serverScriptVersion)
        })
    end)
end

--- @param url string
--- @param callback function    (number/nil),   (any/nil),  (string/nil)
function NMLibrary.getRequest(url, callback)
    NMLibrary.request('GET', url, nil, nil, callback)
end

--- @param url string
--- @param data table
--- @param callback function    (number/nil),   (any/nil),  (string/nil)
function NMLibrary.postRequest(url, data, headers, callback)
    NMLibrary.request('POST', url, data, headers, callback)
end

--- @param method string
--- @param url string
--- @param data table    GET 
--- @param headers table     GET 
--- @param callback function    (number/nil),   (any/nil),  (string/nil)
function NMLibrary.request(method, url, data, headers, callback)
    --    ,   .
    local requestArguments = { headers = headers }

    if method ~= 'GET' then
        requestArguments.headers['Content-Type'] = 'application/json'
        requestArguments.data = encodeJson(data)
    end

    local thread = effil.thread(
        function (method, url, requestArguments)
            --  ,        .
            local requests = require 'requests'

            --  ,  .
            local success, response = pcall(requests.request, method, url, requestArguments)

            --      .
            if success then
                --  ,   .
                response.json, response.xml = nil, nil
                return true, response
            else
                return false, response
            end
        end
    )(method, url, requestArguments)

    --  MoonLoader      ,
    --       callback.
    lua_thread.create(function()
        while true do
            local status, exception = thread:status()

            --  -    -  .
            if exception then
                return callback(nil, nil, exception)
            end

            --     -  .
            if status == 'completed' or status == 'canceled' then
                local success, response = thread:get()

                if success then
                    return callback(response.status_code, decodeJson(response.text) or response.text, nil)
                else
                    return callback(nil, nil, response)
                end
            end

            wait(10)
        end
    end)
end


-- ==  ==
--- @param configDirPath string
--- @param configFilePath string
--- @param defaultConfig table
--- @return table  { isSuccess: boolean, error?: string, config?: table }
function NMLibrary.reloadConfiguration(configDirPath, configFilePath, defaultConfig)
    NMLibrary.createConfigDirectory(configDirPath)

    sendDebug('[reloadConfiguration]  : ' .. tostring(configFilePath) .. '.')
    local config = inicfg.load(defaultConfig, configFilePath)
    sendDebug('[reloadConfiguration]    : ' .. tostring(encodeJson(config)) .. ' (' .. tostring(configFilePath) .. ').')

    if config == nil then
        sendDebug('[reloadConfiguration] -      : ' .. tostring(configFilePath) .. '.')
        return {
            isSuccess = false,
            error = '-      .'
        }
    end

    local saveConfigurationResult = NMLibrary.saveConfiguration(configDirPath, configFilePath, config)
    if not saveConfigurationResult or not saveConfigurationResult.isSuccess then
        return {
            isSuccess = false,
            error = (saveConfigurationResult or { error = '-      .' }).error
        }
    end

    sendDebug('[reloadConfiguration]  : ' .. tostring(encodeJson(config)) .. ' (' .. tostring(configFilePath) .. ').')
    return {
        isSuccess = true,
        config = config
    }
end

--- @param configFilePath string
--- @param configFilePath string
--- @param config table
--- @return table  { isSuccess: boolean, error?: string }
function NMLibrary.saveConfiguration(configDirPath, configFilePath, config)
    NMLibrary.createConfigDirectory(configDirPath)

    sendDebug('[saveConfiguration]  : ' .. tostring(configFilePath) .. ' (' .. tostring(encodeJson(config)) .. ').')
    if not doesFileExist(configFilePath) then
        sendDebug('[saveConfiguration]   : ' .. tostring(configFilePath) .. '.')
        local file = io.open(configFilePath, 'w')

        if not file then
            sendDebug('[saveConfiguration] -       : ' .. tostring(configFilePath) .. '.')
            return {
                isSuccess = false,
                error = '-         .'
            }
        end

        sendDebug('[saveConfiguration]   : ' .. tostring(configFilePath) .. '.')
        file:write('')

        sendDebug('[saveConfiguration]   : ' .. tostring(configFilePath) .. '.')
        file:close()

        sendDebug('[saveConfiguration]   : ' .. tostring(configFilePath) .. '.')
    end

    if not inicfg.save(config, configFilePath) then
        sendDebug('[saveConfiguration] -      : ' .. tostring(configFilePath) .. ' (' .. tostring(encodeJson(config)) .. ').')
        return {
            isSuccess = false,
            error = '-      .'
        }
    end

    sendDebug('[saveConfiguration]  : ' .. tostring(configFilePath) .. ' (' .. tostring(encodeJson(config)) .. ').')
    return { isSuccess = true }
end

--- @param configDirPath string
function NMLibrary.createConfigDirectory(configDirPath)
    sendDebug('[createConfigDirectory]  : ' .. tostring(configDirPath) .. '.')
    createDirectory(configDirPath)
    sendDebug('[createConfigDirectory]  : ' .. tostring(configDirPath) .. '.')
end


-- == SAMP ==
--- @author https://www.blast.hk/members/140618/
--- @param maxDistance number
--- @return any,number car, id
function NMLibrary.getClosestCar(maxDistance)
    local playerX, playerY, playerZ = getCharCoordinates(PLAYER_PED)

    local minDistance = 9999
    local carId = -1
    local carHandle

    for _, car in ipairs(getAllVehicles()) do
        local success, id = sampGetVehicleIdByCarHandle(car)
        if not success then goto nextCar end

        local carX, carY, carZ = getCarCoordinates(car)
        local distance = math.sqrt((carX - playerX) ^ 2 + (carY - playerY) ^ 2 + (carZ - playerZ) ^ 2)

        if distance < minDistance then
            minDistance = distance
            carId = id
            carHandle = car
        end

        ::nextCar::
    end

    if minDistance > maxDistance then
        return nil, -1
    end

    return carHandle, carId
end

---  SAMP   .
--- @param text string
function NMLibrary.stripColors(text)
    return select(1, text:gsub('{%x+}', ''))
end

function NMLibrary.getPlayerID()
    local _, playerID = sampGetPlayerIdByCharHandle(PLAYER_PED)
    return playerID
end

function NMLibrary.getPlayerUsername()
    return sampGetPlayerNickname(NMLibrary.getPlayerID())
end

---  ID ,    .
--- @return number ID  -1
function NMLibrary.getTargetID()
    local success, targetPed = getCharPlayerIsTargeting(PLAYER_HANDLE)
    if not success then return -1 end

    local _, targetID = sampGetPlayerIdByCharHandle(targetPed)
    return targetID
end


-- ==  ==
---   ,      .
--- @param text string
--- @author chapo <https://www.blast.hk/members/112329>
function NMLibrary.len(text)
    return text:gsub('[\128-\191]', ''):len()
end

--- ,      .
--- @param text string
--- @param prefix string
function NMLibrary.startsWith(text, prefix)
    text = string.sub(text, 1, #prefix)
    return text == prefix
end

---    .
--- @param text string
--- @param delimiter string
function NMLibrary.split(text, delimiter)
    local result = {}

    for match in (text .. delimiter):gmatch('(.-)' .. delimiter) do
        table.insert(result, match)
    end

    return result
end

---        .
--- @param text string
function NMLibrary.trim(text)
    return select(1, text:gsub('^%s*(.-)%s*$', '%1'))
end

---  ,     "..",    .
--- @param text string
--- @param maxLength number
function NMLibrary.overlap(text, maxLength)
    if #text <= maxLength then
        return text
    end

    --   {maxLength}    "..".
    return text:sub(1, maxLength) .. '..'
end

---      .
--- @param amount number
function NMLibrary.toMoneyFormat(amount)
    return NMLibrary.trim(
        tostring(amount)
            :reverse()
            :gsub('%d%d%d', '%1 ')
            :reverse()
    )
end


-- ==  ==
--- @param tab table
--- @return table,table  { [index] = key }  { [key] = true }
function NMLibrary.tableKeys(tab)
    local keys = {}
    local key2Bool = {}

    if not tab then
        return keys, key2Bool
    end

    for key, _ in pairs(tab) do
        table.insert(keys, key)
        key2Bool[key] = true
    end

    return keys, key2Bool
end

--- @param tab table
--- @return table  { [index] = value }
function NMLibrary.tableValues(tab)
    local values = {}

    for _, value in pairs(tab) do
        tab.insert(values, value)
    end

    return values
end

---      .
--- @param tab table
--- @return any   nil
function NMLibrary.findKey(tab, targetValue)
    for key, value in pairs(tab) do
        if value == targetValue then
            return key
        end
    end

    return nil
end

---     .
--- @param tab table
--- @return number   nil
function NMLibrary.findIndex(tab, targetValue)
    for i, value in ipairs(tab) do
        if value == targetValue then
            return i
        end
    end

    return nil
end

---     .
--- @param tab table
function NMLibrary.tableSize(tab)
    local size = 0

    for _ in pairs(tab) do size = size + 1 end

    return size
end


-- ==    ==
--- @param dangerousCode function
--- @param exceptionHandler function
function NMLibrary.try(dangerousCode, exceptionHandler)
    local success, exception = pcall(dangerousCode)
    if success then return end

    exceptionHandler(exception)
end

---      .
NMLibrary.ru = encoding.UTF8


return NMLibrary